@@ -1,6 +1,7 @@
 from __future__ import annotations
 import asyncio
 from concurrent.futures import ThreadPoolExecutor, as_completed
+import anyio.to_thread
 from fast_depends import inject
 from fast_depends.core import CallModel
 from fast_depends.dependencies import model
@@ -13,6 +14,8 @@ import tqdm
 import tqdm.asyncio as tqdm_asyncio
 import typing
 from loguru import logger
+import anyio
+from functools import partial
 
 # from functools import up
 from copy import deepcopy
@@ -88,7 +91,7 @@ class Output:
         return type(value).__name__ == expected_type
 
     def __repr__(self):
-        _key_types = ', '.join(f"{k}: {v}" for k, v in self._types.items())
+        _key_types = ", ".join(f"{k}: {v}" for k, v in self._types.items())
         return f"{self.__class__.__name__}({_key_types})"
 
     def as_dict(self) -> dict[str, typing.Any]:
@@ -191,6 +194,9 @@ class Pipeline:
 
         input_spec = self._get_function_input_spec(fn)
 
+        if input_spec.get(_ANONYMOUS) is not None:
+            logger.warning(f"Anonymous type found in function {fn.__name__} parameter ")
+
         for name, _type in input_spec.items():
             _type_reference[_type].add(name)
             self._add_slot(name, _type)
@@ -228,6 +234,7 @@ class Pipeline:
 
         if name in self._slots:
             if self._slots[name] != _data_type_str:
+                logger.debug((self._slots[name], _data_type_str))
                 raise ValueError(
                     f"Slot {name} already exists with " f"type {self._slots[name]}"
                 )
@@ -239,8 +246,13 @@ class Pipeline:
             setattr(self._output, str(name), self._defaults[name])
 
     def _get_function_input_spec(self, fn: IComponent):
+        if hasattr(fn, "__wrapped__"):
+            sig = inspect.signature(getattr(fn, "__wrapped__"))
+        else:
+            sig = inspect.signature(fn)
+
         input_spec = {}
-        sig = inspect.signature(fn)
+        # sig = inspect.signature(_fn)
 
         for name, _type in sig.parameters.items():
             if _type.default is not inspect.Parameter.empty:
@@ -316,6 +328,43 @@ class Pipeline:
 
             if inspect.iscoroutinefunction(fn):
                 fn_output = await fn(**kwargs)
+            else:
+                fn_output = await anyio.to_thread.run_sync(partial(fn, **kwargs))
+
+            if not isinstance(fn_output, tuple):
+                assert len(output_spec) == 1
+                fn_output = (fn_output,)
+
+            for _output, _name in zip(fn_output, output_spec.keys()):
+                setattr(output, str(_name), _output)
+
+        output.lock()
+
+        return output
+
+    def run_sync(self, **inputs):
+        output = self._output.copy()
+
+        output._data = inputs
+
+        for fn in self._pipeline:
+            input_spec, output_spec = self._func_metadata[fn]
+            kwargs = {}
+            for name, _type in input_spec.items():
+                if name in inputs:
+                    kwargs[name] = deepcopy(inputs[name])
+                elif hasattr(output, name):
+                    kwargs[name] = deepcopy(getattr(output, name))
+                else:
+                    raise ValueError(
+                        f"Missing input {name} for " f"function {fn.__name__}"
+                    )
+
+            if self._debug:
+                logger.debug(f"Running {fn.__name__} with inputs {kwargs}")
+
+            if inspect.iscoroutinefunction(fn):
+                fn_output = asyncio.run(fn(**kwargs))
             else:
                 fn_output = fn(**kwargs)
 
@@ -396,7 +445,7 @@ class Pipeline:
             futures = []
             for input in inputs:
                 pl = self.copy()
-                future = _executor.submit(pl, **input)
+                future = _executor.submit(pl.run_sync, **input)
                 futures.append(future)
 
             if preserve_order:
@@ -467,10 +516,10 @@ class Pipeline:
                     display_progress=display_progress,
                 )
             )
-            
+
     def __repr__(self) -> str:
-        _slot_types = ', '.join(f"{k}: {v}" for k, v in self._slots.items())
-        return f'Pipeline({_slot_types})'
+        _slot_types = ", ".join(f"{k}: {v}" for k, v in self._slots.items())
+        return f"Pipeline({_slot_types})"
 
 
 def pipeline(*fns: IComponent, **defaults):
