@@ -5,6 +5,7 @@ import dataclasses
 import glob
 import typing
 from collections.abc import Iterable, MutableMapping
+from jsonpath_rust_bindings import Finder
 
 import farmhash
 import jsonlines
@@ -13,6 +14,7 @@ import slugify
 from box import Box
 from pydantic import BaseModel
 from pydantic.main import IncEx
+from ._functional import filter_nulls, try_chain
 
 # Type aliases
 DictLike = typing.Union[dict, "DotDict"]
@@ -25,6 +27,7 @@ TVal = str | int | float | bool | None
 
 T = typing.TypeVar("T", bound=dict | list)
 
+
 def sort_object_keys(obj: T) -> T:
     if isinstance(obj, dict):
         return {k: sort_object_keys(v) for k, v in sorted(obj.items())}
@@ -33,6 +36,67 @@ def sort_object_keys(obj: T) -> T:
     return obj
 
 
+def sort_if_sortable(obj: T) -> T:
+    try:
+        return sorted(obj)
+    except TypeError:
+        return obj
+
+
+_fallback_serializer = try_chain(
+    lambda data: farmhash.hash64(
+        orjson.dumps(sort_object_keys(filter_nulls(data)), default=serialize)
+    ),
+    lambda data: farmhash.hash64(str(hash(data))),
+)
+
+
+def hash_serialize(data: typing.Any) -> typing.Any:
+    if hasattr(data, "__hash__") and getattr(data, "__hash__"):
+        return hash(data)
+    elif isinstance(data, BaseModel):
+        return farmhash.hash64(
+            orjson.dumps(
+                sort_object_keys(filter_nulls(data.model_dump(mode="json"))),
+                default=hash_serialize,
+            )
+        )
+    elif isinstance(data, dict):
+        return farmhash.hash64(
+            orjson.dumps(sort_object_keys(filter_nulls(data)), default=hash_serialize)
+        )
+    elif isinstance(data, list):
+        return farmhash.hash64(
+            orjson.dumps(sort_if_sortable(filter_nulls(data)), default=hash_serialize)
+        )
+    elif isinstance(data, (tuple, set)):
+        return hash_serialize(sort_if_sortable(filter_nulls(list(data))))
+    else:
+        return _fallback_serializer(data)
+
+
+def serialize(data: typing.Any) -> typing.Any:
+    """
+    Serialize
+    """
+    if hasattr(data, "__hash__") and getattr(data, "__hash__"):
+        return hash(data)
+    elif isinstance(data, dict):
+        return farmhash.hash64(
+            orjson.dumps(sort_object_keys(filter_nulls(data)), default=serialize)
+        )
+    elif isinstance(data, list):
+        return farmhash.hash64(
+            orjson.dumps(sort_if_sortable(filter_nulls(data)), default=serialize)
+        )
+    elif isinstance(data, (tuple, set)):
+        return serialize(sort_if_sortable(filter_nulls(list(data))))
+    elif isinstance(data, bytes):
+        return hash(data)
+    else:
+        return str(data)
+
+
 def object_farmhash(obj: BaseModel | typing.Any, exclude_keys: IncEx = None) -> int:
     """
     Generate a consistent FarmHash64 hash from an object (include Pydantic models)
@@ -41,29 +105,10 @@ def object_farmhash(obj: BaseModel | typing.Any, exclude_keys: IncEx = None) ->
     :return: 64-bit FarmHash as an integer
     """
 
-    def serialize(data: typing.Any) -> typing.Any:
-        if isinstance(data, BaseModel):
-            return serialize(
-                sort_object_keys(
-                    data.model_dump(
-                        mode="json",
-                        exclude_none=True,
-                        exclude=exclude_keys
-                    )
-                )
-            )
-        elif isinstance(data, dict):
-            return {k: serialize(v) for k, v in sorted(data.items())}
-        elif isinstance(data, list):
-            return [serialize(item) for item in sorted(data)]
-        elif isinstance(data, tuple):
-            return tuple(serialize(item) for item in data)
-        elif isinstance(data, set):
-            return frozenset(serialize(item) for item in sorted(data))
-        else:
-            return str(data)
+    if isinstance(obj, BaseModel):
+        obj = obj.model_dump(mode="json", exclude_none=True, exclude=exclude_keys)
 
-    serialized_data = serialize(obj)
+    serialized_data = hash_serialize(obj)
     data_str = str(serialized_data)
     return farmhash.hash64(data_str)
 
@@ -502,34 +547,6 @@ def recursive_convert_lists(obj: dict | list | typing.Any) -> dict | list | typi
 
 # Index and deindex operations
 
-# def index_object(obj: dict, keys_as_tuples: bool = False) -> dict:
-#     """
-#     Create an indexed representation of a nested object.
-
-#     Args:
-#         obj (dict): Object to index.
-#         keys_as_tuples (bool, optional): Whether to use tuples as keys. Defaults to False.
-
-#     Returns:
-#         dict: Indexed representation of the object.
-#     """
-#     def _index_object(obj, path=None):
-#         if path is None:
-#             path = []
-#         if isinstance(obj, dict):
-#             for k, v in obj.items():
-#                 yield from _index_object(v, path + [k])
-#         elif isinstance(obj, list):
-#             for i, v in enumerate(obj):
-#                 yield from _index_object(v, path + [f'[{i}]'])
-#         else:
-#             yield tuple(path), obj
-
-#     return {
-#         path_tuple_to_string(path) if not keys_as_tuples else path: value
-#         for path, value in _index_object(obj)
-#     }
-
 
 def index_object(obj: dict, keys_as_tuples: bool = False) -> dict:
     def _index_object(obj, path=None):
@@ -553,33 +570,6 @@ def index_object(obj: dict, keys_as_tuples: bool = False) -> dict:
     return indexed
 
 
-# def index_object(obj: typing.Any, prefix: str = '') -> dict[str, typing.Any]:
-#     if isinstance(obj, dict):
-#         return {f"{prefix}{key}": index_object(value, f"{prefix}{key}.")
-#                 for key, value in obj.items() if value is not None}
-#     elif isinstance(obj, list):
-#         return {f"{prefix}[{i}]": index_object(value, f"{prefix}[{i}].")
-#                 for i, value in enumerate(obj) if value is not None}
-#     else:
-#         return obj
-
-# def deindex_object(obj: dict[str, TVal]) -> TReturn:
-#     """
-#     Convert an indexed object back to its nested form.
-
-#     Args:
-#         obj (dict[str, TVal]): Indexed object to convert.
-
-
-#     Returns:
-#         dict[str, typing.Any]: Nested form of the object.
-#     """
-#     # return _deindex_object(
-#     #     {path_string_to_tuple(key): value for key, value in obj.items()}
-#     # )
-#     return _deindex_object(
-#         {path_string_to_tuple(key): value for key, value in obj.items()}
-#     )
 def deindex_object(obj: dict[str, TVal]) -> TReturn:
     def _set_nested(result, keys, value):
         current = result
@@ -710,24 +700,6 @@ def path_string_to_tuple(path: str) -> tuple:
 
 TReturn = dict[TKey, typing.Any] | list[typing.Any] | str | int | float | bool | None
 
-# def _deindex_object(obj: dict[tuple[TKey, ...], TVal]) -> TReturn:
-#     """
-#     Convert an indexed object back to its nested form.
-
-#     Args:
-#         obj (dict[tuple[TKey, ...], TVal]): Indexed object to convert.
-
-#     Returns:
-#         TReturn: Nested form of the object, which can be:
-#             - A dictionary with string or int keys and any values
-#             - A list of any values
-#             - A primitive value (str, int, float, bool, or None)
-#     """
-#     result: dict[TKey, typing.Any] = {}
-#     for key, value in obj.items():
-#         _set_nested_dict(result, key, value)
-#     return _numeric_dicts_to_list(result)
-
 
 def _deindex_object(obj: dict[tuple[TKey, ...], TVal]) -> TReturn:
     result: dict[TKey, typing.Any] = {}
@@ -860,8 +832,6 @@ def convert_index_object_to_dict(index):
 
 
 # DotDict implementation
-
-
 class DotDict(MutableMapping):
     def __init__(self, init_dict: DictLike | None = None, **kwargs: typing.Any):
         self.__dict__["_data"] = {}
@@ -918,84 +888,13 @@ class DotDict(MutableMapping):
         return {k: _to_dict(v) for k, v in self._data.items()}
 
 
-# class DotDict(MutableMapping):
-#     """
-#     A dictionary that supports attribute-style access (dot notation).
+def find(json: dict, query: str):
+    return Finder(json).find(query)
 
-#     This class extends the standard python dictionary to allow accessing
-#     and setting items using dot notation (e.g., `d.key` instead of `d['key']`).
 
-#     Args:
-#         init_dict (DictLike | None, optional): Initial dictionary to populate the DotDict. Defaults to None.
-#         **kwargs: Additional key-value pairs to add to the DotDict.
+def find_non_empty(json: dict, query: str):
+    return Finder(json).find_non_empty(query)
 
-#     Example:
-#         >>> d = DotDict({'a': 1, 'b': {'c': 2}})
-#         >>> d.a
-#         1
-#         >>> d.b.c
-#         2
-#         >>> d.d = 3
-#         >>> d['d']
-#         3
-#     """
-
-#     def __init__(self, init_dict: DictLike | None = None, **kwargs: typing.Any):
-#         if init_dict:
-#             super().update(init_dict)
-#         super().update(kwargs)
-
-#     def __getitem__(self, key: str) -> typing.Any:
-#         return self.__dict__[key]
-
-#     def __setitem__(self, key: str, value: typing.Any) -> None:
-#         self.__dict__[key] = value
-
-#     def __delitem__(self, key: str) -> None:
-#         del self.__dict__[key]
-
-#     def __iter__(self) -> typing.Iterator[str]:
-#         return iter(self.__dict__)
-
-#     def __len__(self) -> int:
-#         return len(self.__dict__)
-
-#     def __repr__(self) -> str:
-#         if len(self) > 0:
-#             return f"<{type(self).__name__}: {', '.join(sorted(repr(k) for k in self.keys()))}>"
-#         else:
-#             return f"<{type(self).__name__}>"
-
-#     def get(self, key: str, default: typing.Any = None) -> typing.Any:
-#         """
-#         Get a value from the DotDict.
-
-#         Args:
-#             key (str): The key to retrieve.
-#             default (typing.Any, optional): Default value to return if the key is not found. Defaults to None.
-
-#         Returns:
-#             typing.Any: The value associated with the key, or the default value if not found.
-#         """
-#         return super().get(key, default)
-
-#     def copy(self) -> "DotDict":
-#         """
-#         Create a shallow copy of the DotDict.
-
-#         Returns:
-#             DotDict: A new DotDict instance with copied key-value pairs.
-#         """
-#         return type(self)(self.__dict__.copy())
-
-#     def to_dict(self) -> dict:
-#         """
-#         Convert the DotDict to a regular dictionary.
-
-#         Returns:
-#             dict: A regular dictionary representation of the DotDict.
-#         """
-#         return typing.cast(dict, as_nested_dict(self, dct_class=dict))
 
 # Helper functions for DotDict
 
