@@ -1,9 +1,95 @@
 import typing
 import re
 from collections.abc import Mapping, Sequence
+from loguru import logger
+
+from functools import reduce
+from typing import Callable, TypeVar, overload
+
+T = TypeVar("T")
+T1 = TypeVar("T1")
+T2 = TypeVar("T2")
+T3 = TypeVar("T3")
+T4 = TypeVar("T4")
+T5 = TypeVar("T5")
+T6 = TypeVar("T6")
 
 
-T = typing.TypeVar("T")
+@overload
+def simple_pipeline(f1: Callable[[T], T1]) -> Callable[[T], T1]: ...
+
+
+@overload
+def simple_pipeline(
+    f1: Callable[[T], T1], f2: Callable[[T1], T2]
+) -> Callable[[T], T2]: ...
+
+
+@overload
+def simple_pipeline(
+    f1: Callable[[T], T1], f2: Callable[[T1], T2], f3: Callable[[T2], T3]
+) -> Callable[[T], T3]: ...
+
+
+@overload
+def simple_pipeline(
+    f1: Callable[[T], T1],
+    f2: Callable[[T1], T2],
+    f3: Callable[[T2], T3],
+    f4: Callable[[T3], T4],
+) -> Callable[[T], T4]: ...
+
+
+@overload
+def simple_pipeline(
+    f1: Callable[[T], T1],
+    f2: Callable[[T1], T2],
+    f3: Callable[[T2], T3],
+    f4: Callable[[T3], T4],
+    f5: Callable[[T4], T5],
+) -> Callable[[T], T5]: ...
+
+
+@overload
+def simple_pipeline(*functions: Callable) -> Callable: ...
+
+
+def simple_pipeline(*functions: Callable) -> Callable:
+    """Create a simple_pipeline function from a sequence of callables.
+
+    The returned function's input type matches the first function's input type,
+    and the output type matches the last function's output type.
+
+    Args:
+        *functions: Variable number of callables to compose left-to-right
+
+    Returns:
+        A single callable that applies all functions in sequence
+
+    Example:
+        >>> process = simple_pipeline(str.lower, str.strip, len)
+        >>> process("  HELLO  ")
+        5
+
+        >>> # With type inference
+        >>> transform = simple_pipeline(
+        ...     lambda x: x.split(),      # str -> list[str]
+        ...     lambda x: [w.upper() for w in x],  # list[str] -> list[str]
+        ...     len                        # list[str] -> int
+        ... )
+        >>> transform("hello world")
+        2
+    """
+    if not functions:
+        return lambda x: x
+
+    def piped(initial):
+        return reduce(lambda x, f: f(x), functions, initial)
+
+    # Copy some metadata from the first function
+    piped.__name__ = f"simple_pipeline({', '.join(f.__name__ for f in functions if hasattr(f, '__name__'))})"
+
+    return piped
 
 
 def try_chain(
@@ -37,16 +123,16 @@ def deep_attribute_get(
 ]:
     segments = re.split(r"\.(?![^\[]*\])", path)
     if debug:
-        print(f"Path segments: {segments}")
+        logger.debug(f"Path segments: {segments}")
     result = _traverse(obj, segments, default, debug, "", return_paths)
     flattened = _flatten_and_filter(result, return_paths)
 
     if return_paths:
         filtered = [(value, path) for value, path in flattened if value is not None]
-        return filtered[0] if len(filtered) == 1 else filtered
+        return filtered[0] if len(filtered) == 1 else (filtered or None)
     else:
         filtered = [value for value in flattened if value is not None]
-        return filtered[0] if len(filtered) == 1 else filtered
+        return filtered[0] if len(filtered) == 1 else None
 
 
 def _traverse(
@@ -252,14 +338,10 @@ def set_defaults(__base: dict | None = None, **kwargs):
     return __base
 
 
-# def filter_nulls(obj):
-#     if isinstance(obj, dict):
-#         return {k: filter_nulls(v) for k, v in obj.items() if v}
-#     elif isinstance(obj, list):
-#         return [filter_nulls(v) for v in obj if v]
-#     else:
-#         return obj
-def filter_nulls(obj):
+T = typing.TypeVar("T")
+
+
+def filter_nulls(obj: T) -> T:
     if isinstance(obj, dict):
         return {
             k: v
@@ -270,3 +352,23 @@ def filter_nulls(obj):
         return [v for v in (filter_nulls(v) for v in obj) if v not in (None, {}, [])]
     else:
         return obj
+
+
+def dict_without_keys(
+    d: dict,
+    keys: list[str],
+) -> dict:
+    return {k: v for k, v in d.items() if k not in keys} if d else {}
+
+
+def function_args_dict(
+    function: typing.Callable,
+    args: tuple = (),
+    kwargs: dict = {},
+):
+    import inspect
+
+    sig = inspect.signature(function)
+    bound_args = sig.bind(*args, **kwargs)
+    bound_args.apply_defaults()
+    return bound_args.arguments
